@@ -1,39 +1,50 @@
-import subprocess
+from __future__ import absolute_import
+import copy
 import json
 import logging
-import os
 import re
-import copy
+from typing import Any, AnyStr, Dict, List, Text, Union
+from .utils import docker_windows_path_adjust
+import six
+from six import u
 
-from typing import Any, AnyStr, Union, Text, Dict, List
-import schema_salad.validate as validate
-import schema_salad.ref_resolver
-
-from .utils import aslist, get_feature
-from .errors import WorkflowException
 from . import sandboxjs
-from . import docker
+from .errors import WorkflowException
+from .utils import bytes2str_in_dicts
 
 _logger = logging.getLogger("cwltool")
 
+
 def jshead(engineConfig, rootvars):
     # type: (List[Text], Dict[Text, Any]) -> Text
+
+    # make sure all the byte strings are converted
+    # to str in `rootvars` dict.
+    # TODO: need to make sure the `rootvars dict`
+    # contains no bytes type in the first place.
+    if six.PY3:
+        rootvars = bytes2str_in_dicts(rootvars)  # type: ignore
+
     return u"\n".join(engineConfig + [u"var %s = %s;" % (k, json.dumps(v, indent=4)) for k, v in rootvars.items()])
 
 
+# decode all raw strings to unicode
 seg_symbol = r"""\w+"""
 seg_single = r"""\['([^']|\\')+'\]"""
 seg_double = r"""\["([^"]|\\")+"\]"""
 seg_index = r"""\[[0-9]+\]"""
 segments = r"(\.%s|%s|%s|%s)" % (seg_symbol, seg_single, seg_double, seg_index)
-segment_re = re.compile(segments, flags=re.UNICODE)
-param_re = re.compile(r"\((%s)%s*\)$" % (seg_symbol, segments), flags=re.UNICODE)
+segment_re = re.compile(u(segments), flags=re.UNICODE)
+param_str = r"\((%s)%s*\)$" % (seg_symbol, segments)
+param_re = re.compile(u(param_str), flags=re.UNICODE)
+
+JSON = Union[Dict[Any, Any], List[Any], Text, int, float, bool, None]
 
-JSON = Union[Dict[Any,Any], List[Any], Text, int, long, float, bool, None]
 
 class SubstitutionError(Exception):
     pass
 
+
 def scanner(scan):  # type: (Text) -> List[int]
     DEFAULT = 0
     DOLLAR = 1
@@ -58,13 +69,13 @@ def scanner(scan):  # type: (Text) -> List[int]
         elif state == BACKSLASH:
             stack.pop()
             if stack[-1] == DEFAULT:
-                return [i-1, i+1]
+                return [i - 1, i + 1]
         elif state == DOLLAR:
             if c == '(':
-                start = i-1
+                start = i - 1
                 stack.append(PAREN)
             elif c == '{':
-                start = i-1
+                start = i - 1
                 stack.append(BRACE)
             else:
                 stack.pop()
@@ -74,7 +85,7 @@ def scanner(scan):  # type: (Text) -> List[int]
             elif c == ')':
                 stack.pop()
                 if stack[-1] == DOLLAR:
-                    return [start, i+1]
+                    return [start, i + 1]
             elif c == "'":
                 stack.append(SINGLE_QUOTE)
             elif c == '"':
@@ -85,7 +96,7 @@ def scanner(scan):  # type: (Text) -> List[int]
             elif c == '}':
                 stack.pop()
                 if stack[-1] == DOLLAR:
-                    return [start, i+1]
+                    return [start, i + 1]
             elif c == "'":
                 stack.append(SINGLE_QUOTE)
             elif c == '"':
@@ -103,14 +114,16 @@ def scanner(scan):  # type: (Text) -> List[int]
         i += 1
 
     if len(stack) > 1:
-        raise SubstitutionError("Substitution error, unfinished block starting at position {}: {}".format(start, scan[start:]))
+        raise SubstitutionError(
+            "Substitution error, unfinished block starting at position {}: {}".format(start, scan[start:]))
     else:
         return None
 
+
 def next_seg(remain, obj):  # type: (Text, Any) -> Any
     if remain:
         m = segment_re.match(remain)
-        key = None  # type: Union[str, int]
+        key = None  # type: Union[Text, int]
         if m.group(0)[0] == '.':
             key = m.group(0)[1:]
         elif m.group(0)[1] in ("'", '"'):
@@ -127,7 +140,7 @@ def next_seg(remain, obj):  # type: (Text, Any) -> Any
             try:
                 key = int(m.group(0)[1:-1])
             except ValueError as v:
-                raise WorkflowException(unicode(v))
+                raise WorkflowException(u(str(v)))
             if not isinstance(obj, list):
                 raise WorkflowException(" is a %s, cannot index on int '%s'" % (type(obj).__name__, key))
             if key >= len(obj):
@@ -139,10 +152,13 @@ def next_seg(remain, obj):  # type: (Text, Any) -> Any
     else:
         return obj
 
+
 def evaluator(ex, jslib, obj, fullJS=False, timeout=None, debug=False):
     # type: (Text, Text, Dict[Text, Any], bool, int, bool) -> JSON
     m = param_re.match(ex)
     if m:
+        if m.end(1)+1 == len(ex) and m.group(1) == "null":
+            return None
         try:
             return next_seg(m.group(0)[m.end(1) - m.start(0):-1], obj[m.group(1)])
         except Exception as w:
@@ -150,7 +166,10 @@ def evaluator(ex, jslib, obj, fullJS=False, timeout=None, debug=False):
     elif fullJS:
         return sandboxjs.execjs(ex, jslib, timeout=timeout, debug=debug)
     else:
-        raise sandboxjs.JavascriptException("Syntax error in parameter reference '%s' or used Javascript code without specifying InlineJavascriptRequirement.", ex)
+        raise sandboxjs.JavascriptException(
+            "Syntax error in parameter reference '%s' or used Javascript code without specifying InlineJavascriptRequirement.",
+            ex)
+
 
 def interpolate(scan, rootvars,
                 timeout=None, fullJS=None, jslib="", debug=False):
@@ -162,7 +181,7 @@ def interpolate(scan, rootvars,
         parts.append(scan[0:w[0]])
 
         if scan[w[0]] == '$':
-            e = evaluator(scan[w[0]+1:w[1]], jslib, rootvars, fullJS=fullJS,
+            e = evaluator(scan[w[0] + 1:w[1]], jslib, rootvars, fullJS=fullJS,
                           timeout=timeout, debug=debug)
             if w[0] == 0 and w[1] == len(scan):
                 return e
@@ -171,7 +190,7 @@ def interpolate(scan, rootvars,
                 leaf = leaf[1:-1]
             parts.append(leaf)
         elif scan[w[0]] == '\\':
-            e = scan[w[1]-1]
+            e = scan[w[1] - 1]
             parts.append(e)
 
         scan = scan[w[1]:]
@@ -179,18 +198,19 @@ def interpolate(scan, rootvars,
     parts.append(scan)
     return ''.join(parts)
 
+
 def do_eval(ex, jobinput, requirements, outdir, tmpdir, resources,
             context=None, pull_image=True, timeout=None, debug=False):
     # type: (Union[dict, AnyStr], Dict[Text, Union[Dict, List, Text]], List[Dict[Text, Any]], Text, Text, Dict[Text, Union[int, Text]], Any, bool, int, bool) -> Any
 
     runtime = copy.copy(resources)
-    runtime["tmpdir"] = tmpdir
-    runtime["outdir"] = outdir
+    runtime["tmpdir"] = docker_windows_path_adjust(tmpdir)
+    runtime["outdir"] = docker_windows_path_adjust(outdir)
 
     rootvars = {
         u"inputs": jobinput,
         u"self": context,
-        u"runtime": runtime }
+        u"runtime": runtime}
 
     if isinstance(ex, (str, Text)):
         fullJS = False
