@@ -1,26 +1,30 @@
-import subprocess
+from __future__ import absolute_import
+import functools
 import io
-import os
-import tempfile
-import glob
 import json
 import logging
-import sys
-import requests
-from . import docker
-from .process import get_feature, empty_subtree, stageFiles
-from .errors import WorkflowException
+import os
+import re
 import shutil
 import stat
-import re
+import subprocess
+import sys
+import tempfile
+from io import open
+from typing import (IO, Any, Callable, Dict, Iterable, List, MutableMapping, Text,
+                    Tuple, Union, cast)
+
 import shellescape
-import string
-from .docker_uid import docker_vm_uid
+
+from .utils import copytree_with_merge, docker_windows_path_adjust, onWindows
+from . import docker
 from .builder import Builder
-from typing import (Any, Callable, Union, Iterable, Mapping, MutableMapping,
-        IO, cast, Text, Tuple)
+from .docker_id import docker_vm_id
+from .errors import WorkflowException
 from .pathmapper import PathMapper
-import functools
+from .process import (UnsupportedRequirement, empty_subtree, get_feature,
+                      stageFiles)
+from .utils import bytes2str_in_dicts
 
 _logger = logging.getLogger("cwltool")
 
@@ -34,6 +38,7 @@ python "run_job.py" "job.json"
 
 PYTHON_RUN_SCRIPT = """
 import json
+import os
 import sys
 import subprocess
 
@@ -42,6 +47,7 @@ with open(sys.argv[1], "r") as f:
     commands = popen_description["commands"]
     cwd = popen_description["cwd"]
     env = popen_description["env"]
+    env["PATH"] = os.environ.get("PATH")
     stdin_path = popen_description["stdin_path"]
     stdout_path = popen_description["stdout_path"]
     stderr_path = popen_description["stderr_path"]
@@ -68,7 +74,7 @@ with open(sys.argv[1], "r") as f:
     if sp.stdin:
         sp.stdin.close()
     rcode = sp.wait()
-    if isinstance(stdin, file):
+    if stdin is not subprocess.PIPE:
         stdin.close()
     if stdout is not sys.stderr:
         stdout.close()
@@ -83,6 +89,7 @@ def deref_links(outputs):  # type: (Any) -> None
         if outputs.get("class") == "File":
             st = os.lstat(outputs["path"])
             if stat.S_ISLNK(st.st_mode):
+                outputs["basename"] = os.path.basename(outputs["path"])
                 outputs["path"] = os.readlink(outputs["path"])
         else:
             for v in outputs.values():
@@ -91,8 +98,26 @@ def deref_links(outputs):  # type: (Any) -> None
         for v in outputs:
             deref_links(v)
 
-class CommandLineJob(object):
+def relink_initialworkdir(pathmapper, inplace_update=False):
+    # type: (PathMapper, bool) -> None
+    for src, vol in pathmapper.items():
+        if not vol.staged:
+            continue
+        if vol.type in ("File", "Directory") or (inplace_update and
+                                                 vol.type in ("WritableFile", "WritableDirectory")):
+            if os.path.islink(vol.target) or os.path.isfile(vol.target):
+                os.remove(vol.target)
+            elif os.path.isdir(vol.target):
+                shutil.rmtree(vol.target)
+            if onWindows():
+                if vol.type in ("File", "WritableFile"):
+                    shutil.copy(vol.resolved,vol.target)
+                elif vol.type in ("Directory", "WritableDirectory"):
+                    copytree_with_merge(vol.resolved, vol.target)
+            else:
+                os.symlink(vol.resolved, vol.target)
 
+class JobBase(object):
     def __init__(self):  # type: () -> None
         self.builder = None  # type: Builder
         self.joborder = None  # type: Dict[Text, Union[Dict[Text, Any], List, Text]]
@@ -107,6 +132,7 @@ class CommandLineJob(object):
         self.name = None  # type: Text
         self.command_line = None  # type: List[Text]
         self.pathmapper = None  # type: PathMapper
+        self.generatemapper = None  # type: PathMapper
         self.collect_outputs = None  # type: Union[Callable[[Any], Any], functools.partial[Any]]
         self.output_callback = None  # type: Callable[[Any, Any], Any]
         self.outdir = None  # type: Text
@@ -114,100 +140,31 @@ class CommandLineJob(object):
         self.environment = None  # type: MutableMapping[Text, Text]
         self.generatefiles = None  # type: Dict[Text, Union[List[Dict[Text, Text]], Dict[Text, Text], Text]]
         self.stagedir = None  # type: Text
+        self.inplace_update = None  # type: bool
 
-    def run(self, dry_run=False, pull_image=True, rm_container=True,
-            rm_tmpdir=True, move_outputs="move", **kwargs):
-        # type: (bool, bool, bool, bool, Text, **Any) -> Union[Tuple[Text, Dict[None, None]], None]
+    def _setup(self):  # type: () -> None
         if not os.path.exists(self.outdir):
             os.makedirs(self.outdir)
 
-        #with open(os.path.join(outdir, "cwl.input.json"), "w") as fp:
-        #    json.dump(self.joborder, fp)
-
-        runtime = []  # type: List[Text]
-
-        (docker_req, docker_is_req) = get_feature(self, "DockerRequirement")
-
         for knownfile in self.pathmapper.files():
             p = self.pathmapper.mapper(knownfile)
-            if p.type == "File" and not os.path.isfile(p[0]):
+            if p.type == "File" and not os.path.isfile(p[0]) and p.staged:
                 raise WorkflowException(
                     u"Input file %s (at %s) not found or is not a regular "
                     "file." % (knownfile, self.pathmapper.mapper(knownfile)[0]))
 
-        img_id = None
-        env = None  # type: Union[MutableMapping[Text, Text], MutableMapping[str, str]]
-        if docker_req and kwargs.get("use_container") is not False:
-            env = os.environ
-            img_id = docker.get_from_requirements(docker_req, docker_is_req, pull_image)
-        elif kwargs.get("default_container", None) is not None:
-            env = os.environ
-            img_id = kwargs.get("default_container")
+        if self.generatefiles["listing"]:
+            self.generatemapper = PathMapper(cast(List[Any], self.generatefiles["listing"]),
+                                             self.outdir, self.outdir, separateDirs=False)
+            _logger.debug(u"[job %s] initial work dir %s", self.name,
+                          json.dumps({p: self.generatemapper.mapper(p) for p in self.generatemapper.files()}, indent=4))
 
-        if docker_is_req and img_id is None:
-            raise WorkflowException("Docker is required for running this tool.")
-
-        if img_id:
-            runtime = ["docker", "run", "-i"]
-            for src in self.pathmapper.files():
-                vol = self.pathmapper.mapper(src)
-                if vol.type == "File":
-                    runtime.append(u"--volume=%s:%s:ro" % (vol.resolved, vol.target))
-                if vol.type == "CreateFile":
-                    createtmp = os.path.join(self.stagedir, os.path.basename(vol.target))
-                    with open(createtmp, "w") as f:
-                        f.write(vol.resolved.encode("utf-8"))
-                    runtime.append(u"--volume=%s:%s:ro" % (createtmp, vol.target))
-            runtime.append(u"--volume=%s:%s:rw" % (os.path.realpath(self.outdir), self.builder.outdir))
-            runtime.append(u"--volume=%s:%s:rw" % (os.path.realpath(self.tmpdir), "/tmp"))
-            runtime.append(u"--workdir=%s" % (self.builder.outdir))
-            runtime.append("--read-only=true")
-
-            if kwargs.get("custom_net", None) is not None:
-                runtime.append("--net={0}".format(kwargs.get("custom_net")))
-            elif kwargs.get("disable_net", None):
-                runtime.append("--net=none")
-
-            if self.stdout:
-                runtime.append("--log-driver=none")
-
-            euid = docker_vm_uid() or os.geteuid()
-
-            if kwargs.get("no_match_user",None) is False:
-                runtime.append(u"--user=%s" % (euid))
-
-            if rm_container:
-                runtime.append("--rm")
-
-            runtime.append("--env=TMPDIR=/tmp")
-
-            # spec currently says "HOME must be set to the designated output
-            # directory." but spec might change to designated temp directory.
-            # runtime.append("--env=HOME=/tmp")
-            runtime.append("--env=HOME=%s" % self.builder.outdir)
-
-            for t,v in self.environment.items():
-                runtime.append(u"--env=%s=%s" % (t, v))
-
-            runtime.append(img_id)
-        else:
-            env = self.environment
-            if not os.path.exists(self.tmpdir):
-                os.makedirs(self.tmpdir)
-            vars_to_preserve = kwargs.get("preserve_environment")
-            if kwargs.get("preserve_entire_environment"):
-                vars_to_preserve = os.environ
-            if vars_to_preserve is not None:
-                for key, value in os.environ.items():
-                    if key in vars_to_preserve and key not in env:
-                        env[key] = value
-            env["HOME"] = self.outdir
-            env["TMPDIR"] = self.tmpdir
-
-            stageFiles(self.pathmapper, os.symlink)
+    def _execute(self, runtime, env, rm_tmpdir=True, move_outputs="move"):
+        # type: (List[Text], MutableMapping[Text, Text], bool, Text) -> None
 
         scr, _ = get_feature(self, "ShellCommandRequirement")
 
+        shouldquote = None  # type: Callable[[Any], Any]
         if scr:
             shouldquote = lambda x: False
         else:
@@ -216,32 +173,15 @@ class CommandLineJob(object):
         _logger.info(u"[job %s] %s$ %s%s%s%s",
                      self.name,
                      self.outdir,
-                     " \\\n    ".join([shellescape.quote(Text(arg)) if shouldquote(Text(arg)) else Text(arg) for arg in (runtime + self.command_line)]),
+                     " \\\n    ".join([shellescape.quote(Text(arg)) if shouldquote(Text(arg)) else Text(arg) for arg in
+                                       (runtime + self.command_line)]),
                      u' < %s' % self.stdin if self.stdin else '',
                      u' > %s' % os.path.join(self.outdir, self.stdout) if self.stdout else '',
                      u' 2> %s' % os.path.join(self.outdir, self.stderr) if self.stderr else '')
 
-        if dry_run:
-            return (self.outdir, {})
-
         outputs = {}  # type: Dict[Text,Text]
 
         try:
-            if self.generatefiles["listing"]:
-                generatemapper = PathMapper([self.generatefiles], self.outdir,
-                                            self.outdir, separateDirs=False)
-                _logger.debug(u"[job %s] initial work dir %s", self.name,
-                              json.dumps({p: generatemapper.mapper(p) for p in generatemapper.files()}, indent=4))
-
-                def linkoutdir(src, tgt):
-                    # Need to make the link to the staged file (may be inside
-                    # the container)
-                    for _, item in self.pathmapper.items():
-                        if src == item.resolved:
-                            os.symlink(item.target, tgt)
-                            break
-                stageFiles(generatemapper, linkoutdir)
-
             stdin_path = None
             if self.stdin:
                 stdin_path = self.pathmapper.reversemap(self.stdin)[1]
@@ -262,15 +202,19 @@ class CommandLineJob(object):
                     os.makedirs(dn)
                 stdout_path = absout
 
-            build_job_script = self.builder.build_job_script  # type: Callable[[List[str]], Text]
+            commands = [Text(x).encode('utf-8') for x in runtime + self.command_line]
+            job_script_contents = None  # type: Text
+            builder = getattr(self, "builder", None)  # type: Builder
+            if builder is not None:
+                job_script_contents = builder.build_job_script(commands)
             rcode = _job_popen(
-                [Text(x).encode('utf-8') for x in runtime + self.command_line],
+                commands,
                 stdin_path=stdin_path,
                 stdout_path=stdout_path,
                 stderr_path=stderr_path,
                 env=env,
                 cwd=self.outdir,
-                build_job_script=build_job_script,
+                job_script_contents=job_script_contents,
             )
 
             if self.successCodes and rcode in self.successCodes:
@@ -285,15 +229,10 @@ class CommandLineJob(object):
                 processStatus = "permanentFail"
 
             if self.generatefiles["listing"]:
-                def linkoutdir(src, tgt):
-                    # Need to make the link to the staged file (may be inside
-                    # the container)
-                    if os.path.islink(tgt):
-                        os.remove(tgt)
-                        os.symlink(src, tgt)
-                stageFiles(generatemapper, linkoutdir, ignoreWritable=True)
+                relink_initialworkdir(self.generatemapper, inplace_update=self.inplace_update)
 
             outputs = self.collect_outputs(self.outdir)
+            outputs = bytes2str_in_dicts(outputs)  # type: ignore
 
         except OSError as e:
             if e.errno == 2:
@@ -312,7 +251,7 @@ class CommandLineJob(object):
             processStatus = "permanentFail"
 
         if processStatus != "success":
-            _logger.warn(u"[job %s] completed %s", self.name, processStatus)
+            _logger.warning(u"[job %s] completed %s", self.name, processStatus)
         else:
             _logger.info(u"[job %s] completed %s", self.name, processStatus)
 
@@ -329,27 +268,165 @@ class CommandLineJob(object):
             _logger.debug(u"[job %s] Removing temporary directory %s", self.name, self.tmpdir)
             shutil.rmtree(self.tmpdir, True)
 
-        if move_outputs == "move" and empty_subtree(self.outdir):
-            _logger.debug(u"[job %s] Removing empty output directory %s", self.name, self.outdir)
-            shutil.rmtree(self.outdir, True)
+
+class CommandLineJob(JobBase):
+
+    def run(self, pull_image=True, rm_container=True,
+            rm_tmpdir=True, move_outputs="move", **kwargs):
+        # type: (bool, bool, bool, Text, **Any) -> None
+
+        self._setup()
+
+        env = self.environment
+        if not os.path.exists(self.tmpdir):
+            os.makedirs(self.tmpdir)
+        vars_to_preserve = kwargs.get("preserve_environment")
+        if kwargs.get("preserve_entire_environment"):
+            vars_to_preserve = os.environ
+        if vars_to_preserve is not None:
+            for key, value in os.environ.items():
+                if key in vars_to_preserve and key not in env:
+                    # On Windows, subprocess env can't handle unicode.
+                    env[key] = str(value) if onWindows() else value
+        env["HOME"] = str(self.outdir) if onWindows() else self.outdir
+        env["TMPDIR"] = str(self.tmpdir) if onWindows() else self.tmpdir
+        if "PATH" not in env:
+            env["PATH"] = str(os.environ["PATH"]) if onWindows() else os.environ["PATH"]
+
+        stageFiles(self.pathmapper, ignoreWritable=True, symLink=True)
+        if self.generatemapper:
+            stageFiles(self.generatemapper, ignoreWritable=self.inplace_update, symLink=True)
+            relink_initialworkdir(self.generatemapper, inplace_update=self.inplace_update)
+
+        self._execute([], env, rm_tmpdir=rm_tmpdir, move_outputs=move_outputs)
+
+
+class DockerCommandLineJob(JobBase):
+
+    def add_volumes(self, pathmapper, runtime, stage_output):
+        # type: (PathMapper, List[Text], bool) -> None
+
+        host_outdir = self.outdir
+        container_outdir = self.builder.outdir
+        for src, vol in pathmapper.items():
+            if not vol.staged:
+                continue
+            if stage_output:
+                containertgt = container_outdir + vol.target[len(host_outdir):]
+            else:
+                containertgt = vol.target
+            if vol.type in ("File", "Directory"):
+                if not vol.resolved.startswith("_:"):
+                    runtime.append(u"--volume=%s:%s:ro" % (docker_windows_path_adjust(vol.resolved), docker_windows_path_adjust(containertgt)))
+            elif vol.type == "WritableFile":
+                if self.inplace_update:
+                    runtime.append(u"--volume=%s:%s:rw" % (docker_windows_path_adjust(vol.resolved), docker_windows_path_adjust(containertgt)))
+                else:
+                    shutil.copy(vol.resolved, vol.target)
+            elif vol.type == "WritableDirectory":
+                if vol.resolved.startswith("_:"):
+                    os.makedirs(vol.target, 0o0755)
+                else:
+                    if self.inplace_update:
+                        runtime.append(u"--volume=%s:%s:rw" % (docker_windows_path_adjust(vol.resolved), docker_windows_path_adjust(containertgt)))
+                    else:
+                        shutil.copytree(vol.resolved, vol.target)
+            elif vol.type == "CreateFile":
+                createtmp = os.path.join(host_outdir, os.path.basename(vol.target))
+                with open(createtmp, "wb") as f:
+                    f.write(vol.resolved.encode("utf-8"))
+                runtime.append(u"--volume=%s:%s:ro" % (docker_windows_path_adjust(createtmp), docker_windows_path_adjust(vol.target)))
+
+
+    def run(self, pull_image=True, rm_container=True,
+            rm_tmpdir=True, move_outputs="move", **kwargs):
+        # type: (bool, bool, bool, Text, **Any) -> None
+
+        (docker_req, docker_is_req) = get_feature(self, "DockerRequirement")
+
+        img_id = None
+        env = None  # type: MutableMapping[Text, Text]
+        try:
+            env = cast(MutableMapping[Text, Text], os.environ)
+            if docker_req and kwargs.get("use_container"):
+                img_id = docker.get_from_requirements(docker_req, True, pull_image)
+            if img_id is None:
+                if self.builder.find_default_container:
+                    default_container = self.builder.find_default_container()
+                    if default_container:
+                        img_id = default_container
+                        env = cast(MutableMapping[Text, Text], os.environ)
+
+            if docker_req and img_id is None and kwargs.get("use_container"):
+                raise Exception("Docker image not available")
+        except Exception as e:
+            _logger.debug("Docker error", exc_info=True)
+            if docker_is_req:
+                raise UnsupportedRequirement(
+                    "Docker is required to run this tool: %s" % e)
+            else:
+                raise WorkflowException(
+                    "Docker is not available for this tool, try --no-container"
+                    " to disable Docker: %s" % e)
+
+        self._setup()
+
+        runtime = [u"docker", u"run", u"-i"]
+
+        runtime.append(u"--volume=%s:%s:rw" % (docker_windows_path_adjust(os.path.realpath(self.outdir)), self.builder.outdir))
+        runtime.append(u"--volume=%s:%s:rw" % (docker_windows_path_adjust(os.path.realpath(self.tmpdir)), "/tmp"))
+
+        self.add_volumes(self.pathmapper, runtime, False)
+        if self.generatemapper:
+            self.add_volumes(self.generatemapper, runtime, True)
+
+        runtime.append(u"--workdir=%s" % (docker_windows_path_adjust(self.builder.outdir)))
+        runtime.append(u"--read-only=true")
+
+        if kwargs.get("custom_net", None) is not None:
+            runtime.append(u"--net={0}".format(kwargs.get("custom_net")))
+        elif kwargs.get("disable_net", None):
+            runtime.append(u"--net=none")
+
+        if self.stdout:
+            runtime.append("--log-driver=none")
+
+        euid, egid = docker_vm_id()
+        if not onWindows():  # MS Windows does not have getuid() or geteuid() functions
+            euid, egid = euid or os.geteuid(), egid or os.getgid()
+
+        if kwargs.get("no_match_user", None) is False and (euid, egid) != (None, None):
+            runtime.append(u"--user=%d:%d" % (euid, egid))
+
+        if rm_container:
+            runtime.append(u"--rm")
+
+        runtime.append(u"--env=TMPDIR=/tmp")
+
+        # spec currently says "HOME must be set to the designated output
+        # directory." but spec might change to designated temp directory.
+        # runtime.append("--env=HOME=/tmp")
+        runtime.append(u"--env=HOME=%s" % self.builder.outdir)
+
+        for t, v in self.environment.items():
+            runtime.append(u"--env=%s=%s" % (t, v))
+
+        runtime.append(img_id)
+
+        self._execute(runtime, env, rm_tmpdir=rm_tmpdir, move_outputs=move_outputs)
 
 
 def _job_popen(
-    commands,  # type: List[str]
-    stdin_path,  # type: Text
-    stdout_path,  # type: Text
-    stderr_path,  # type: Text
-    env,  # type: Union[MutableMapping[Text, Text], MutableMapping[str, str]]
-    cwd,  # type: Text
-    job_dir=None,  # type: Text
-    build_job_script=None,  # type: Callable[[List[str]], Text]
+        commands,  # type: List[bytes]
+        stdin_path,  # type: Text
+        stdout_path,  # type: Text
+        stderr_path,  # type: Text
+        env,  # type: Union[MutableMapping[Text, Text], MutableMapping[str, str]]
+        cwd,  # type: Text
+        job_dir=None,  # type: Text
+        job_script_contents=None,  # type: Text
 ):
     # type: (...) -> int
-
-    job_script_contents = None  # type: Text
-    if build_job_script:
-        job_script_contents = build_job_script(commands)
-
     if not job_script_contents and not FORCE_SHELLED_POPEN:
 
         stdin = None  # type: Union[IO[Any], int]
@@ -373,7 +450,7 @@ def _job_popen(
 
         sp = subprocess.Popen(commands,
                               shell=False,
-                              close_fds=True,
+                              close_fds=not onWindows(),
                               stdin=stdin,
                               stdout=stdout,
                               stderr=stderr,
@@ -385,7 +462,7 @@ def _job_popen(
 
         rcode = sp.wait()
 
-        if isinstance(stdin, file):
+        if isinstance(stdin, io.IOBase):
             stdin.close()
 
         if stdout is not sys.stderr:
@@ -403,8 +480,8 @@ def _job_popen(
             job_script_contents = SHELL_COMMAND_TEMPLATE
 
         env_copy = {}
+        key = None  # type: Any
         for key in env:
-            key = key.encode("utf-8")
             env_copy[key] = env[key]
 
         job_description = dict(
@@ -415,21 +492,21 @@ def _job_popen(
             stderr_path=stderr_path,
             stdin_path=stdin_path,
         )
-        with open(os.path.join(job_dir, "job.json"), "w") as f:
+        with open(os.path.join(job_dir, "job.json"), "wb") as f:
             json.dump(job_description, f)
         try:
             job_script = os.path.join(job_dir, "run_job.bash")
-            with open(job_script, "w") as f:
-                f.write(job_script_contents)
+            with open(job_script, "wb") as f:
+                f.write(job_script_contents.encode('utf-8'))
             job_run = os.path.join(job_dir, "run_job.py")
-            with open(job_run, "w") as f:
+            with open(job_run, "wb") as f:
                 f.write(PYTHON_RUN_SCRIPT)
             sp = subprocess.Popen(
                 ["bash", job_script.encode("utf-8")],
                 shell=False,
                 cwd=job_dir,
-                stdout=subprocess.PIPE,
-                stderr=subprocess.PIPE,
+                stdout=sys.stderr,  # The nested script will output the paths to the correct files if they need
+                stderr=sys.stderr,  # to be captured. Else just write everything to stderr (same as above).
                 stdin=subprocess.PIPE,
             )
             if sp.stdin:
